#include <iostream>
#include <fstream>
#include <sstream>
#include <map>
#include <set>
#include <algorithm>

#include <dai/util.h>
#include <dai/properties.h>

#include <dai/daialg.h>
#include <dai/alldai.h>
#include <dai/bbp.h>
#include <dai/clamp_test.h>

namespace dai {

using namespace std;

const char *ClampTest::Name = "CLAMP_TEST";

void ClampTest::construct() {
    _beliefsV.clear(); _beliefsV.reserve(nrVars());
    for( size_t i = 0; i < nrVars(); i++ )
        _beliefsV.push_back( Factor(var(i),1.0).normalized() );

    _beliefsF.clear(); _beliefsF.reserve(nrFactors());
    for( size_t I = 0; I < nrFactors(); I++ ) {
        Factor f = factor(I);
        f.fill(1); f.normalize();
        _beliefsF.push_back(f);
    }

    _maxdiff = 0;
    _iters = 0;
}

typedef ClampTest::Properties::ClampType ClampType;

double ClampTest::run() {
    assert(props.clamp==ClampType::CLAMP_VAR || props.clamp==ClampType::CLAMP_FACTOR);
    bool clampingVar = (props.clamp==ClampType::CLAMP_VAR);

    // run base model
    InfAlg *ia = newInfAlgFromString(props.infalg, fg());
    if(typeid(*ia) == typeid(BP)) {
        static_cast<BP*>(ia)->recordSentMessages=1;
    }
    ia->init();
    ia->run();
    _iters = ia->Iterations();

    // split base model according to var=state and var!=state
    size_t i = props.var;
    size_t xi = props.state;
    vector<size_t> xis;
    xis.push_back(xi);
    if(clampingVar) {
        foreach(size_t xi, xis) { assert(/*0<=xi &&*/ xi<var(i).states()); }
    } else {
        foreach(size_t xI, xis) { assert(/*0<=xI &&*/ xI<factor(i).states()); }
    }

    // compute complement of 'xis'
    vector<size_t> cmp_xis=complement(xis, clampingVar ? var(i).states() : factor(i).states());

    // XXX could do this more efficiently with a nesting version of
    // saveProbs/undoProbs
    Real lz; vector<Factor> b;
    InfAlg *ia_c = ia->clone();
    Real cmp_lz; vector<Factor> cmp_b;
    InfAlg *cmp_ia_c = ia->clone();
    if(clampingVar) {
        clampVar(ia_c->fg(), i, xis);
        clampVar(cmp_ia_c->fg(), i, cmp_xis);

        ia_c->init(var(i));
        cmp_ia_c->init(var(i));
    } else {
        clampFactor(ia_c->fg(), i, xis);
        clampFactor(cmp_ia_c->fg(), i, cmp_xis);

        ia_c->init(factor(i).vars());
        cmp_ia_c->init(factor(i).vars());
    }

    ia_c->run();
    cmp_ia_c->run();
    _iters += ia_c->Iterations() + cmp_ia_c->Iterations();

    lz = ia_c->logZ();
    cmp_lz = cmp_ia_c->logZ();

    _logZ = logSumExp(lz, cmp_lz);

    Real p = unSoftMax(lz,cmp_lz);
    Real cmp_p = 1-p;

    foreach_size_t(i, fg().nrVars()) {
        _beliefsV[i] =
            ia_c->beliefV(i).normalized()*p+
            cmp_ia_c->beliefV(i).normalized()*cmp_p;
    }
    foreach_size_t(I, fg().nrFactors()) {
        _beliefsF[I] =
            ia_c->beliefF(I).normalized()*p+
            cmp_ia_c->beliefF(I).normalized()*cmp_p;
    }

    _maxdiff = max(ia_c->maxDiff(), cmp_ia_c->maxDiff());

    vector<size_t> state;
    if(needGibbsState(props.bbp_cfn)) {
        state = getGibbsState(*ia,2*ia->Iterations());
    }

    if(props.run_bbp_test) {
        double err = numericBBPTest(*ia, &state, props.bbp_props, props.bbp_cfn, /*h*/ 1.0e-5);
        cerr << "Error from numericBBPTest: " << err << endl;
    }
    if(props.clamp_plot_file.length()>0) {
        cerr << "Writing clamp plot data" << endl;
        makeClampPlot(*ia,clampingVar,i,xi,&state,props.bbp_cfn,props.clamp_plot_file.c_str());
        cerr << "Done writing BBP plot data" << endl;
    }

    // write stuff to auxfile here
    if(props.auxfile.length()>0) {
        // run BBP here
        BBP bbp(ia,props.bbp_props);
        initBBPCostFnAdj(bbp, *ia, props.bbp_cfn, &state);
        bbp.run();

        ofstream os(props.auxfile.c_str(), ios_base::out|ios_base::trunc);
        // what to write to file:
        // adjoint at [variable/factor],state using bbp_cfn
        // estimated marginal of state for original InfAlg run
        // improved estimated marginal of state ('p' above) from clamped logZs
        os << "# ADJOINT\tMARG1\tMARG2\tCFN1\tCFN2\tDISTL1V\tDISTL1F\n";
        os << (clampingVar ? bbp.adj_psi_V(i)[xi] : bbp.adj_psi_F(i)[xi]) << "\t";
        os << (clampingVar ? ia->beliefV(i)[xi] : ia->beliefF(i)[xi]) << "\t";
        os << p << "\t";
        Real cfn1 = getCostFn(*ia, props.bbp_cfn, &state);
        Real cfn2 = getCostFn(*this, props.bbp_cfn, &state);
        os << cfn1 << "\t" << cfn2 << "\t";
        double distV=0,distF=0;
        for(size_t j=0; j<nrVars(); j++) {
            distV += dist(beliefV(j), ia->beliefV(j), Prob::DISTL1);
        }
        for(size_t J=0; J<nrFactors(); J++) {
            distF += dist(beliefF(J), ia->beliefF(J), Prob::DISTL1);
        }
        os << distV << "\t" << distF << endl;
        os.close();
    }

    if(props.lr_info_file.length()>0) {
        outputLRInfo(props.lr_info_file);
    }

    delete ia_c; delete cmp_ia_c;
    delete ia;
    
    return _maxdiff;
}


void makeClampPlot(const InfAlg& bp, bool clampingVar,
                   size_t i, size_t xi,
                   vector<size_t> *gibbsState,
                   bbp_cfn_t cfn, const char* outfile)
{
    double tiny=1.0e-7;
    size_t num_samples=500;
    if(bp.beliefV(i)[xi]<tiny || abs(bp.beliefV(i)[xi]-1)<tiny) {
        cerr << "Warning: makeClampPlot: extreme marginal for variable " << i << " state " << xi << endl;
    }
    ofstream os(outfile, ios_base::out|ios_base::trunc);
      
    for(size_t n=0; n<num_samples; n++) {
        double c=((double)n)/(num_samples-1);
        InfAlg *bp_prb = bp.clone();

        if(clampingVar) {
            Prob psi_V_prb(bp.fg().var(i).states(),1.0-c);
            psi_V_prb[xi] = 1.0;
            assert(bp_prb->fg().nbV(i).size()>0);
            size_t I = bp_prb->fg().nbV(i)[0]; // modify the first factor in list of neighbours of i
            bp_prb->fg().factor(I) *= Factor(bp.fg().var(i), psi_V_prb);
            bp_prb->init(bp.fg().var(i));
        } else {
            Prob psi_F_prb(bp.fg().factor(i).states(),1.0-c);
            psi_F_prb[xi] = 1.0;
            bp_prb->fg().factor(i).p() *= psi_F_prb;
            bp_prb->init(bp.fg().factor(i).vars());
        }
        bp_prb->run();
        Real cf = getCostFn(*bp_prb, cfn, gibbsState);
        os << c << "\t" << cf << endl;
    }
      
    os.close();
}

Factor ClampTest::belief( const VarSet &ns ) const {
    if( ns.size() == 1 )
        return belief( *(ns.begin()) );
    else {
        size_t I;
        for( I = 0; I < nrFactors(); I++ )
            if( factor(I).vars() >> ns )
                break;
        assert( I != nrFactors() );
        return beliefF(I).marginal(ns);
    }
}

void ClampTest::outputLRInfo(string outfile) {
    // run inference algorithm
    InfAlg *ia = newInfAlgFromString(props.infalg, fg());
    if(typeid(*ia) == typeid(BP)) {
        static_cast<BP*>(ia)->recordSentMessages=1;
    }
    ia->init();
    ia->run();

    // open output file
    ofstream os(outfile.c_str(), ios_base::out|ios_base::trunc);
    // write column names to file
    os << "# VAR1 ST1 VAR2 ST2 DIST JOINT COV" << endl;
    bool use_cond=1;
    vector<vector<Prob> > b_ij_cond;
    vector<vector<int> > distances;
    // for each variable i
    foreach_size_t(i, nrVars()) {
        b_ij_cond.push_back(vector<Prob>());
        // get distances of each other variable
        distances.push_back(distancesV(i));
        // for each state of i
        size_t n_i = var(i).states();
        foreach_size_t(x_i, n_i) {
            if(use_cond) {
                vector<size_t> xis;
                xis.push_back(x_i);

                Real bixi = ia->beliefV(i).p()[x_i];
                // clone ia
                InfAlg *ia_c = ia->clone();
                // clamp i to x_i
                clampVar(ia_c->fg(), i, xis);
                // rerun ia
                ia_c->init(var(i));
                ia_c->run();

                // for each variable j
                foreach_size_t(j, nrVars()) {
                    // initialise b_ij_cond[i][j]
                    b_ij_cond[i].push_back(Prob(var(i).states()*var(j).states(),0.0));
                    // for each state x_j of j
                    foreach_size_t(x_j, var(j).states()) {
                        // compute b_ij^COND(x_i,x_j) (as function of x_j), by multiplying with bixi
                        b_ij_cond[i][j][x_j*n_i+x_i] = ia_c->beliefV(j)[x_j]*bixi;
                    }
                }
            } else {
                // initialise BBP adjoints to calculate all p(x_i,x_j)-p(x_i)p(x_j)
                BBP bbp(ia,props.bbp_props);
                vector<Prob> b1_adj;
                b1_adj.reserve(fg().nrVars());
                foreach_size_t(j, nrVars()) {
                    size_t n_j = var(j).states();
                    Prob delta(n_j,0.0);
                    if(j==i) {
                        delta[x_i] = 1.0;
                    }
                    b1_adj.push_back(delta);
                }
                bbp.init(b1_adj);
                // run BBP
                bbp.run();
                // for each variable j, and state x_j
                foreach_size_t(j, nrVars()) {
                    Prob cov = bbp.adj_psi_V(j);
                    Prob joint = cov + ia->beliefV(j).p()*ia->beliefV(i)[x_i];
                    foreach_size_t(x_j, var(j).states()) {
                        // write to file: distance to j, p(x_i,x_j), p(x_i,x_j)-p(x_i)p(x_j)
                        os << i << "\t" << x_i << "\t" << j << "\t" << x_j << "\t"
                           << distances[i][j] << "\t" << joint[x_j] << "\t" << cov[x_j] << endl;
                    }
                }
            }
        }
    }
    if(use_cond) {
        // loop over pairs of variables
        foreach_size_t(i, nrVars()) {
            foreach_size_t(j, nrVars()) {
                size_t n_i = var(i).states();
                size_t n_j = var(j).states();
                // compute marginals
                //     b_ij_cond[i][j][x_j*n_i+x_i] = ia_c->beliefV(j)[x_j]*bixi;
                Prob joint = b_ij_cond[i][j];
                Prob cov(n_i*n_j,0), b_i_cond(n_i,0), b_j_cond(n_j,0);
                foreach_size_t(x_i, n_i) {
                    foreach_size_t(x_j, n_j) {
                        Real v = joint[x_j*n_i+x_i];
                        b_i_cond[x_i] += v;
                        b_j_cond[x_j] += v;
                    }
                }
                // loop over pairs of states, compute covariance and write to file
                foreach_size_t(x_i, n_i) {
                    foreach_size_t(x_j, n_j) {
                        Real joint_ = joint[x_j*n_i+x_i];
                        Real cov = joint_ - b_i_cond[x_i]*b_j_cond[x_j];
                        os << i << "\t" << x_i << "\t" << j << "\t" << x_j << "\t"
                           << distances[i][j] << "\t" << joint_ << "\t" << cov << endl;
                    }
                }
            }
        }
    }
    // delete ia
    delete ia;
        
    // close output file
    os.close();
}


}


/* {{{ GENERATED CODE: DO NOT EDIT. Created by 
    ./scripts/regenerate-properties include/dai/clamp_test.h src/clamp_test.cpp 
*/
namespace dai {

void ClampTest::Properties::set(const PropertySet &opts)
{
    const std::set<PropertyKey> &keys = opts.allKeys();
    std::set<PropertyKey>::const_iterator i;
    bool die=false;
    for(i=keys.begin(); i!=keys.end(); i++) {
        if(*i == "verbose") continue;
        if(*i == "clamp") continue;
        if(*i == "var") continue;
        if(*i == "state") continue;
        if(*i == "infalg") continue;
        if(*i == "bbp_cfn") continue;
        if(*i == "bbp_props") continue;
        if(*i == "auxfile") continue;
        if(*i == "lr_info_file") continue;
        if(*i == "clamp_plot_file") continue;
        if(*i == "run_bbp_test") continue;
        cerr << "ClampTest: Unknown property " << *i << endl;
        die=true;
    }
    if(die) {
        DAI_THROW(UNKNOWN_PROPERTY_TYPE);
    }
    if(!opts.hasKey("verbose")) {
        cerr << "ClampTest: Missing property \"verbose\" for method \"ClampTest\"" << endl;
        die=true;
    }
    if(!opts.hasKey("clamp")) {
        cerr << "ClampTest: Missing property \"clamp\" for method \"ClampTest\"" << endl;
        die=true;
    }
    if(!opts.hasKey("var")) {
        cerr << "ClampTest: Missing property \"var\" for method \"ClampTest\"" << endl;
        die=true;
    }
    if(!opts.hasKey("state")) {
        cerr << "ClampTest: Missing property \"state\" for method \"ClampTest\"" << endl;
        die=true;
    }
    if(!opts.hasKey("infalg")) {
        cerr << "ClampTest: Missing property \"infalg\" for method \"ClampTest\"" << endl;
        die=true;
    }
    if(!opts.hasKey("bbp_cfn")) {
        cerr << "ClampTest: Missing property \"bbp_cfn\" for method \"ClampTest\"" << endl;
        die=true;
    }
    if(!opts.hasKey("bbp_props")) {
        cerr << "ClampTest: Missing property \"bbp_props\" for method \"ClampTest\"" << endl;
        die=true;
    }
    if(!opts.hasKey("auxfile")) {
        cerr << "ClampTest: Missing property \"auxfile\" for method \"ClampTest\"" << endl;
        die=true;
    }
    if(!opts.hasKey("lr_info_file")) {
        cerr << "ClampTest: Missing property \"lr_info_file\" for method \"ClampTest\"" << endl;
        die=true;
    }
    if(!opts.hasKey("clamp_plot_file")) {
        cerr << "ClampTest: Missing property \"clamp_plot_file\" for method \"ClampTest\"" << endl;
        die=true;
    }
    if(!opts.hasKey("run_bbp_test")) {
        cerr << "ClampTest: Missing property \"run_bbp_test\" for method \"ClampTest\"" << endl;
        die=true;
    }
    if(die) {
        DAI_THROW(NOT_ALL_PROPERTIES_SPECIFIED);
    }
    verbose = opts.getStringAs<size_t>("verbose");
    clamp = opts.getStringAs<ClampType>("clamp");
    var = opts.getStringAs<size_t>("var");
    state = opts.getStringAs<size_t>("state");
    infalg = opts.getStringAs<string>("infalg");
    bbp_cfn = opts.getStringAs<bbp_cfn_t>("bbp_cfn");
    bbp_props = opts.getStringAs<PropertySet>("bbp_props");
    auxfile = opts.getStringAs<string>("auxfile");
    lr_info_file = opts.getStringAs<string>("lr_info_file");
    clamp_plot_file = opts.getStringAs<string>("clamp_plot_file");
    run_bbp_test = opts.getStringAs<bool>("run_bbp_test");
}
PropertySet ClampTest::Properties::get() const {
    PropertySet opts;
    opts.Set("verbose", verbose);
    opts.Set("clamp", clamp);
    opts.Set("var", var);
    opts.Set("state", state);
    opts.Set("infalg", infalg);
    opts.Set("bbp_cfn", bbp_cfn);
    opts.Set("bbp_props", bbp_props);
    opts.Set("auxfile", auxfile);
    opts.Set("lr_info_file", lr_info_file);
    opts.Set("clamp_plot_file", clamp_plot_file);
    opts.Set("run_bbp_test", run_bbp_test);
    return opts;
}
string ClampTest::Properties::toString() const {
    stringstream s(stringstream::out);
    s << "[";
    s << "verbose=" << verbose << ",";
    s << "clamp=" << clamp << ",";
    s << "var=" << var << ",";
    s << "state=" << state << ",";
    s << "infalg=" << infalg << ",";
    s << "bbp_cfn=" << bbp_cfn << ",";
    s << "bbp_props=" << bbp_props << ",";
    s << "auxfile=" << auxfile << ",";
    s << "lr_info_file=" << lr_info_file << ",";
    s << "clamp_plot_file=" << clamp_plot_file << ",";
    s << "run_bbp_test=" << run_bbp_test;
    s << "]";
    return s.str();
}
} // end of namespace dai
/* }}} END OF GENERATED CODE */
